%macro extrr3fl(outlib=,language=,merge=,release=,
                tmplib=WORK,
                where=,whdlm=,tables_only=N);
%*******************************************************************;
%* Copyright(C) 1998 by SAS Institute Inc., Cary, NC USA           *;
%*                                                                 *;
%* Macro: extrr3fl                                                 *;
%*                                                                 *;
%* Usage: %extrr3fl(outlib   = target library,                     *;
%*                  language = language key,                       *;
%*                  merge    = Y|N flag: merge into existing data  *;
%*                             dictionary extract,                 *;
%*                  release  = 3|4|6 SAP R/3 Release,              *;
%*                  tmplib   = temporary library                   *;
%*                  where    = ABAP/4 where clause to subset on    *;
%*                             tables,                             *;
%*                  whdlm    = delimiter for where statement for   *;
%*                             multiple lines,                     *;
%*                  tables_only = Y|N,                             *;
%*                             Y - only extract fields for         *;
%*                                 tables in the tables data set.  *;
%*                                 The where clause specified in   *;
%*                                 where= applies to the SAS data  *;
%*                                 set.                            *;
%*                             N - extract field metadata for all  *;
%*                                 tables                          *;
%*                 );                                              *;
%*                                                                 *;
%*        example:                                                 *;
%*         %let max_secs=1200;   * split R/3 extraction into blocks*;
%*                               * of 20 minutes (20*60=1200)      *;
%*         %extrr3fl(outlib   = R3LIB,                             *;
%*                   language = E,                                 *;
%*                   merge    = N,                                 *;
%*                   release  = 3,                                 *;
%*                   where    = TABNAME BETWEEN 'T001' AND 'T010');*;
%*                                                                 *;
%* Description:                                                    *;
%*   This macro is used for extracting metadata from the R/3 data  *;
%*   dictionary. It extract the fields info.                       *;
%*                                                                 *;
%*******************************************************************;

        %global r3conn &r3conn.C max_secs last_tab last_run;

        %local wh i ifm _stmt;
        %local lwhere1 lwhere2 lwhere3 lwhere4 lwhere5
               lwhere6 lwhere7 lwhere8 lwhere9 lwhere10
               lwhere11 lwhere12 lwhere13 lwhere14 lwhere15
               lwhere16 lwhere17 lwhere18 lwhere19 lwhere20
               lwhere21 lwhere22 lwhere23 lwhere24 lwhere25
               lwhere26 lwhere27 lwhere28 lwhere29 lwhere30;
        %local dsid;

        %let dsid=0;

        %if %eval(&&&r3conn.C>0 and &&&r3conn.C<1000 and
                  %sysfunc(rank(A))=65) %then
            %let ifm=EBCDIC;
        %else %if %eval(&&&r3conn.C>=1000 and &&&r3conn.C<2000 and
                  %sysfunc(rank(A))^=65) %then
            %let ifm=ASCII;
        %else
            %let ifm=;

      DATA &tmplib..VFIELDS / VIEW=&tmplib..VFIELDS;

         n = -1;
         length aa $ 200;
         retain flag 0;
         retain errflg 0;

         if flag eq 0 then do;
            flag = 1;

                             /* The Data Connection */
            n = -1;
            infile &r3conn.D truncover nbyte=n lrecl=22 eof=eof;

            input d1 $ d2 $char200.;

                             /* The Request Connection */
            infile &r3conn.R obs=1000 truncover nbyte=n;
              file &r3conn.R;

            input r1 $ r2 $char200.;
            _r3save_ = _error_;
            ch = symget("&r3conn.H");
            if _error_ then _error_ = _r3save_;
            if ch ne "" then do;
               file &r3conn.R; put "CHANDLE " ch;
               link errtst;
            end;

            l = symget("LANGUAGE");
            if _error_ then _error_ = _r3save_;
            if l ne "" then do;
               l = upcase(l);
               file &r3conn.R; put "CPARM l " l;
               link errtst;

               file &r3conn.R; put "CPICPARM l " l;
               link errtst;
            end;

            f = symget("FUNCTION");
            if _error_ then _error_ = _r3save_;
            if f ne "" then do;
               f = upcase(f);
               file &r3conn.R; put "FMOD " f;
               link errtst;
            end;

            aa = "TABLE FIELDS";
            file &r3conn.R; put aa;
            link errtst;


        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;

            aa = "LENGTH 180";
            file &r3conn.R; put aa;
            link errtst;

        %end;

        %* before R/3 release 4;
        %else %do;

            aa = "LENGTH 120";
            file &r3conn.R; put aa;
            link errtst;

        %end;

            aa = "CPICPARM f FLDDESC2";
            file &r3conn.R; put aa;
            link errtst;

            aa = "WHERESTART";
            file &r3conn.R; put aa;
            link errtst;

        %if %upcase(%str(&tables_only))=Y %then %do;
            /* Ignore where clause in macro variable where.
               The actual where clause is built
               from the data set &outlib..TABLES.

               It also does not use the time slicing logic. */

            do ii=1 to 30;
               aa=symget("LWHERE"||trim(left(put(ii,6.))));

               if aa="" then leave;

               aa="WHERETXT "||aa;
               file &r3conn.R; put aa;
               link errtst;
            end;
            if ii>1 then do;
               aa = "WHERETXT AND";
               file &r3conn.R; put aa;
               link errtst;
            end;
        %end;
        %else %do;

        %if %nrbquote(&where) ne %then %do;

          %let i=1;
          %let wh=%scan(&where,&i,&whdlm);
          %do %while (%nrbquote(&wh) ne );

            aa = "WHERETXT "||"&wh";
            file &r3conn.R; put aa;
            link errtst;

            %let i=%eval(&i+1);
            %let wh=%scan(&where,&i,&whdlm);
          %end;

            aa = "WHERETXT AND";
            file &r3conn.R; put aa;
            link errtst;
        %end;

        %if %datatyp(&max_secs)=NUMERIC and &max_secs>0 %then %do;
            aa = "WHERETXT ( TABNAME GT '"||
                 trim(left(symget('last_tab')))||"' ) AND ";
            file &r3conn.R; put aa;
            link errtst;

            aa = "WHERETXT ( FIELDNAME GT ' ' ) AND ";
            file &r3conn.R; put aa;
            link errtst;
        %end;

        %end;

            * select relevant AS4LOCAL and
            * also enforce order of extract by including key as4vers! ;
            aa = "WHERETXT ( AS4LOCAL EQ 'A' ) AND ( AS4VERS GE '0' ) ";
            file &r3conn.R; put aa;
            link errtst;

            aa = "WHEREEND";
            file &r3conn.R; put aa;
            link errtst;

            file &r3conn.R; put "DHANDLE " d2;
            link errtst;

            aa = "DCONNECT";
            file &r3conn.R; put aa;
            link errtst;

         end;

        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;

         n = 181;
         infile &r3conn.D truncover nbyte=n lrecl=181 eof=eof;

        %end;

        %* before R/3 release 4;
        %else %do;

         n = 121;
         infile &r3conn.D truncover nbyte=n lrecl=121 eof=eof;

        %end;

         length fldfmt                           $  4
                fldtext                          $ 60
                fldafmt keyflag                  $  1
                flddec flddblen fldmlen fldoffs     8
                language                         $  1
                adminfie                         $  1;

        %if %str(&ifm) ne %str() %then %do;
         length flddec0 flddblen0 fldmlen0 $ 6 fldoffs0 $ 4;
        %end;

        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;

         length fldname tabname rollname         $ 30;
         format fldname tabname rollname         $30.;

        %end;

        %* before R/3 release 4;
        %else %do;

         length fldname tabname rollname         $ 10;
         format fldname tabname rollname         $10.;

        %end;

         keep fldfmt fldtext fldname tabname fldafmt keyflag
              flddec flddblen fldmlen fldoffs language adminfie
              rollname;

         format fldfmt                            $4.
                fldtext                           $60.
                fldafmt keyflag                   $1.
                flddec flddblen fldmlen fldoffs    6.
                language                          $1.
                adminfie                          $1.;

         label fldfmt   = "Dictionary Data Type"
               fldtext  = "Short Description"
               fldname  = "Field Name"
               tabname  = "Table Name"
               fldafmt  = "Internal ABAP Type"
               keyflag  = "Primary Key Flag"
               flddec   = "No. of decimal places"
               flddblen = "Internal Field Length"
               fldmlen  = "External Field Length"
               fldoffs  = "Offset"
               language = "Language"
               adminfie = "Include Level"
               rollname = "Data Element (semantic domain)";

         input @;
         input @n test $1. @;

         if test = '00'x then do;
            errflg=1;
            input @1 w1 $ @;
            if w1 =: 'ERROR' or w1 = '+ERROR' then do;
               input @1 aa $char200. @;
               substr(aa, 1, 6) = 'ERROR:';
               file log; put aa;
               if w1 =: 'ERROR' then abort 8;
            end;
            else do;
               file log; put _infile_;
            end;
            delete;
            return;
         end;

        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;
            %if %str(&ifm) eq %str() %then %do;

         input @1 fldfmt       $4.
                  fldtext      $60.
                  flddec    ?? 6.
                  fldname      $30.
                  flddblen  ?? 6.
                  fldafmt      $1.
                  keyflag      $1.
                  fldmlen   ?? 6.
                  fldoffs   ?? 4.
                  tabname      $30.
                  language     $1.
                  adminfie     $1.
                  rollname     $30.;

            %end;
            %else %do;
         input @1 fldfmt    $&ifm.4.
                  fldtext   $&ifm.60.
                  flddec0   $&ifm.6.
                  fldname   $&ifm.30.
                  flddblen0 $&ifm.6.
                  fldafmt   $&ifm.1.
                  keyflag   $&ifm.1.
                  fldmlen0  $&ifm.6.
                  fldoffs0  $&ifm.4.
                  tabname   $&ifm.30.
                  language  $&ifm.1.
                  adminfie  $&ifm.1.
                  rollname  $&ifm.30.;
        flddec=input(flddec0,?? 6.);
        flddblen=input(flddblen0,?? 6.);
        fldmlen=input(fldmlen0,?? 6.);
        fldoffs=input(fldoffs0,?? 4.);
            %end;
        %end;

        %* before R/3 release 4;
        %else %do;
            %if %str(&ifm) eq %str() %then %do;

         input @1 fldfmt       $4.
                  fldtext      $60.
                  flddec    ?? 6.
                  fldname      $10.
                  flddblen  ?? 6.
                  fldafmt      $1.
                  keyflag      $1.
                  fldmlen   ?? 6.
                  fldoffs   ?? 4.
                  tabname      $10.
                  language     $1.
                  adminfie     $1.
                  rollname     $10.;

            %end;
            %else %do;
         input @1 fldfmt    $&ifm.4.
                  fldtext   $&ifm.60.
                  flddec0   $&ifm.6.
                  fldname   $&ifm.10.
                  flddblen0 $&ifm.6.
                  fldafmt   $&ifm.1.
                  keyflag   $&ifm.1.
                  fldmlen0  $&ifm.6.
                  fldoffs0  $&ifm.4.
                  tabname   $&ifm.10.
                  language  $&ifm.1.
                  adminfie  $&ifm.1.
                  rollname  $&ifm.10.;
        flddec=input(flddec0,?? 6.);
        flddblen=input(flddblen0,?? 6.);
        fldmlen=input(fldmlen0,?? 6.);
        fldoffs=input(fldoffs0,?? 4.);
            %end;
        %end;


      return;

      errtst:
         input aa $char200.;
         w1 = scan(aa, 1, " ");
         if w1 = 'OK' then do;
            return;
         end;
         else do;
            errflg=1;
            do while (w1 ne 'OK');
               if w1 =: 'ERROR' or w1 eq '+ERROR' then do;
                  substr(aa, 1, 6) = 'ERROR:';
               end;
               file log; put aa;
               if w1 =: 'ERROR' then abort 8;
               input aa $char200.;
               w1 = scan(aa, 1, " ");
            end;
         end;
      return;
      eof:
         if errflg then abort 8;
         delete;
      return;
      run;
      %if &syserr^=0 %then %goto exitfl;

      %let last_tab=;
      %if %datatyp(&max_secs)=NUMERIC and &max_secs>0 %then %do;
          %let _stmt=
               %str(retain _strttm;)
               %str(if _n_=1 then _strttm=datetime();)
               %str(by tabname notsorted;)
               %str(drop _strttm;)
               %str(if last.tabname then do;)
               %str(   output;)
               %str(   if (datetime()-_strttm) >&max_secs then do;)
               %str(      file log ;)
               %str(      call symput('last_tab',tabname);)
               %str(      call symput('last_run','0');)
               %str(      stop;)
               %str(   end;)
               %str(end;)
               %str(else output;);
      %end;
      %else %let _stmt=;

      /* delete output data set */
      %if %upcase(%str(&merge)) ne %str(Y) %then %do;
          proc datasets lib=&outlib nowarn nolist;
               delete fields;
          quit;
      %end;

      %if %upcase(%str(&tables_only)) eq Y %then %do;
          %local base saswhere;

          %if %sysfunc(exist(&outlib..fields)) eq 1 %then %do;
              %let base=&tmplib..fields;
          %end;
          %else %do;
              %let base=&outlib..fields;
          %end;

          %let saswhere=;

          %if %nrbquote(&where) ne %then %do;

              %let i=1;
              %let wh=%scan(&where,&i,&whdlm);
              %do %while (%nrbquote(&wh) ne );
                  %let saswhere=&saswhere &wh;
                  %let i=%eval(&i+1);
                  %let wh=%scan(&where,&i,&whdlm);
              %end;
          %end;

          %if %nrbquote(&saswhere) ne %then %do;
              %let saswhere=&saswhere AND LANGUAGE="&language";
          %end;
          %else %do;
              %let saswhere=LANGUAGE="&language";
          %end;


          %local rc vn tabname i_where;

          data &tmplib..stables;
               set &outlib..tables;
               where &saswhere;
          run;
          %if &syserr^=0 %then %goto exitfl;

          %let dsid=%sysfunc(open(&tmplib..stables));
          %if %eval(&dsid)>0 %then %do;
              %let vn=%sysfunc(varnum(&dsid,TABNAME));

              %let i_where=1;

              %do %while(%eval(%sysfunc(fetch(&dsid)))=0);
                  %let tabname=%sysfunc(getvarc(&dsid,&vn));

                  %if %eval(&i_where)>1 %then %do;
                      %let lwhere&&i_where=%str(, %'&tabname%');
                  %end;
                  %else %do;
                      %let lwhere&&i_where=%str(TABNAME IN %(%'&tabname%');
                  %end;
                  %let i_where=%eval(&i_where+1);
                  %if %eval(&i_where)=30 %then %do;
                      %let lwhere&&i_where=%str(%));

          proc append base=&base new=&tmplib..vfields force;
          run;
          %if &syserr^=0 %then %goto exitfl;

                      %let i_where=1;
                  %end;
              %end;

              %let rc=%sysfunc(close(&dsid));
              %let dsid=0;

              %if %eval(&i_where)>1 %then %do;
                  %let lwhere&i_where=%str(%));
                  %let i_where=%eval(&i_where+1);
                  %if %eval(&i_where) le 30 %then %do;
                      %let lwhere&i_where=;
                  %end;
          proc append base=&base new=&tmplib..vfields force;
          run;
          %if &syserr^=0 %then %goto exitfl;
              %end;
          %end;

          %if %upcase(%str(&merge)) eq Y and
              %sysfunc(exist(&outlib..fields)) eq 1 %then %do;

          proc sort data=&outlib..fields force;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exitfl;

          proc sort data=&tmplib..fields;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exitfl;

          data &outlib..fields;
               set &tmplib..fields(in=new) &outlib..fields(in=old);
               by language tabname;
               length _lang $ 2 _tab $ 30;
               drop _lang _tab;
               retain _lang _tab;
               if first.tabname and new then do;
                  _lang=language; _tab=tabname;
               end;
               if old then do;
                  if language=_lang and tabname=_tab then delete;
               end;
          run;
          %end;
      %end;


      %else %if %upcase(%str(&merge)) eq %str(Y) and
          %sysfunc(exist(&outlib..fields)) eq 1 %then %do;

          %let last_run=0;

          %do %until (&last_run=1);
              %let last_run=1;

          data &tmplib..fields;
               set &tmplib..vfields;
               &_stmt;
          run;
          %if &syserr^=0 %then %goto exitfl;

          proc sort data=&outlib..fields force;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exitfl;

          proc sort data=&tmplib..fields;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exitfl;

          data &outlib..fields;
               set &tmplib..fields(in=new) &outlib..fields(in=old);
               by language tabname;
               length _lang $ 2 _tab $ 30;
               drop _lang _tab;
               retain _lang _tab;
               if first.tabname and new then do;
                  _lang=language; _tab=tabname;
               end;
               if old then do;
                  if language=_lang and tabname=_tab then delete;
               end;
          run;
          %if &syserr^=0 %then %goto exitfl;

          %end;
      %end;
      %else %do;
          %let last_run=0;

          %do %until (&last_run=1);
              %let last_run=1;

              %if %datatyp(&max_secs)=NUMERIC and
                &max_secs>0 and %str(&last_tab) ne %str() %then %do;

          data &tmplib..fields;
               set &tmplib..vfields;
               &_stmt;
          run;
          %if &syserr^=0 %then %goto exitfl;

          proc append base=&outlib..fields new=&tmplib..fields;
          run;
          %if &syserr^=0 %then %goto exitfl;

              %end;
              %else %do;

          data &outlib..fields(label='R/3 DD: Field Information');
               set &tmplib..vfields;
               &_stmt;
          run;
          %if &syserr^=0 %then %goto exitfl;

              %end;
          %end;
      %end;

      proc sort data=&outlib..fields force;
           by language tabname;
      run;

      proc datasets lib=&outlib nolist;
           modify FIELDS;
              index create TABNAME;
      quit;

%exitfl:
      %if %eval(&dsid)>0 %then %do;
          %let rc=%sysfunc(close(&dsid));
      %end;

%mend;
