%macro extrr3tb(outlib=,language=,merge=,release=,
                tmplib=WORK,
                where=,whdlm=);
%*******************************************************************;
%* Copyright(C) 1998 by SAS Institute Inc., Cary, NC USA           *;
%*                                                                 *;
%* Macro: extrr3tb                                                 *;
%*                                                                 *;
%* Usage: %extrr3tb(outlib   = target library,                     *;
%*                  language = language key,                       *;
%*                  merge    = Y|N flag: merge into existing data  *;
%*                             dictionary extract,                 *;
%*                  release  = 3|4|6 SAP R/3 Release,              *;
%*                  where    = ABAP/4 where clause to subset on    *;
%*                             tables,                             *;
%*                  whdlm    = delimiter for where statement for   *;
%*                             multiple lines                      *;
%*                 );                                              *;
%*                                                                 *;
%*        example:                                                 *;
%*         %extrr3tb(outlib   = R3LIB,                             *;
%*                   language = E,                                 *;
%*                   merge    = N,                                 *;
%*                   conn     = CONN1,                             *;
%*                   release  = 3,                                 *;
%*                   what     = TABLES FIELDS CREATE_VIEW          *;
%*                   where    = TABNAME BETWEEN 'T001' AND 'T010');*;
%*                                                                 *;
%* Description:                                                    *;
%*   This macro is used for extracting metadata from the R/3 data  *;
%*   dictionary. It extracts the table info.                       *;
%*                                                                 *;
%*******************************************************************;

        %global r3conn &r3conn.C;

        %local wh i ifm;

        %if %eval(&&&r3conn.C>0 and &&&r3conn.C<1000 and
                  %sysfunc(rank(A))=65) %then
            %let ifm=EBCDIC;
        %else %if %eval(&&&r3conn.C>=1000 and &&&r3conn.C<2000 and
                  %sysfunc(rank(A))^=65) %then
            %let ifm=ASCII;
        %else
            %let ifm=;


      DATA &tmplib..VTABLES / VIEW=&tmplib..VTABLES;

         n = -1;
         length aa $ 200;
         retain flag 0;
         retain errflg 0;

         if flag eq 0 then do;
            flag = 1;

                             /* The Data Connection */
            n = -1;
            infile &r3conn.D truncover nbyte=n lrecl=22 eof=eof;

            input d1 $ d2 $char200.;

                             /* The Request Connection */
            infile &r3conn.R obs=1000 truncover nbyte=n;
              file &r3conn.R;

            input r1 $ r2 $char200.;
            _r3save_ = _error_;
            ch = symget("&r3conn.H");
            if _error_ then _error_ = _r3save_;
            if ch ne "" then do;
               file &r3conn.R; put "CHANDLE " ch;
               link errtst;
            end;

            l = symget("LANGUAGE");
            if _error_ then _error_ = _r3save_;
            if l ne "" then do;
               l = upcase(l);
               file &r3conn.R; put "CPARM l " l;
               link errtst;

               file &r3conn.R; put "CPICPARM l " l;
               link errtst;
            end;

            f = symget("FUNCTION");
            if _error_ then _error_ = _r3save_;
            if f ne "" then do;
               f = upcase(f);
               file &r3conn.R; put "FMOD " f;
               link errtst;
            end;

            aa = "TABLE TABLES";
            file &r3conn.R; put aa;
            link errtst;

        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;

            aa = "LENGTH 99";
            file &r3conn.R; put aa;
            link errtst;

        %end;

        %* before R/3 release 4;
        %else %do;

            aa = "LENGTH 79";
            file &r3conn.R; put aa;
            link errtst;

        %end;

            aa = "CPICPARM f TABDESC";
            file &r3conn.R; put aa;
            link errtst;

        %* apply the given where clause to the data extraction;
        %if %nrbquote(&where) ne %then %do;

            aa = "WHERESTART";
            file &r3conn.R; put aa;
            link errtst;

          %let i=1;
          %let wh=%scan(&where,&i,&whdlm);
          %do %while (%nrbquote(&wh) ne );

            aa= "WHERETXT "||"&wh";
            file &r3conn.R; put aa;
            link errtst;

            %let i=%eval(&i+1);
            %let wh=%scan(&where,&i,&whdlm);
          %end;

            aa = "WHEREEND";
            file &r3conn.R; put aa;
            link errtst;

        %end;

            file &r3conn.R; put "DHANDLE " d2;
            link errtst;

            aa = "DCONNECT";
            file &r3conn.R; put aa;
            link errtst;

         end;

        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;

         n = 100;
         infile &r3conn.D truncover nbyte=n lrecl=100 eof=eof;

        %end;

        %* before R/3 release 4;
        %else %do;

         n = 80;
         infile &r3conn.D truncover nbyte=n lrecl=80 eof=eof;

        %end;


        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;

         length tabname $ 30 tabclass $ 8 tabtext $ 60 language $ 1;
         format tabname $30. tabclass $8. tabtext $60. language $1.;

        %end;

        %* before R/3 release 4;
        %else %do;

         length tabname $ 10 tabclass $ 8 tabtext $ 60 language $ 1;
         format tabname $10. tabclass $8. tabtext $60. language $1.;

        %end;

         keep tabname tabclass tabtext language;

         label tabname   = 'Table name'
               tabclass  = 'Table type'
               tabtext   = 'Short description'
               language  = 'Language';

         input @n test $1. @;
         if test = '00'x then do;
            errflg=1;
            input @1 w1 $ @;
            if w1 =: 'ERROR' or w1 = '+ERROR' then do;
               input @1 aa $char200. @;
               substr(aa, 1, 6) = 'ERROR:';
               file log; put aa;
               if w1 =: 'ERROR' then abort 8;
            end;
            else do;
               file log; put _infile_;
            end;
            delete;
            return;
         end;

        %* R/3 release 4;
        %if %nrbquote(%substr(&release,1,1))=4 or
            %nrbquote(%substr(&release,1,1))=6 %then %do;

         input @1 tabname  $&ifm.30.
                  tabclass $&ifm.8.
                  tabtext  $&ifm.60.
                  language $&ifm.1.;

        %end;

        %* before R/3 release 4;
        %else %do;

         input @1 tabname  $&ifm.10.
                  tabclass $&ifm.8.
                  tabtext  $&ifm.60.
                  language $&ifm.1.;
        %end;

      return;

      errtst:
         input aa $char200.;
         w1 = scan(aa, 1, " ");
         if w1 = 'OK' then do;
            return;
         end;
         else do;
            errflg=1;
            do while (w1 ne 'OK');
               if w1 =: 'ERROR' or w1 eq '+ERROR' then do;
                  substr(aa, 1, 6) = 'ERROR:';
               end;
               file log; put aa;
               if w1 =: 'ERROR' then abort 8;
               input aa $char200.;
               w1 = scan(aa, 1, " ");
            end;
         end;
      return;
      eof:
         if errflg then abort 8;
         delete;
      return;
      run;
      %if &syserr^=0 %then %goto exittb;

      %if %upcase(%str(&merge)) eq %str(Y) and
          %sysfunc(exist(&outlib..TABLES)) eq 1 %then %do;

          data &tmplib..tables;
               set &tmplib..vtables;
          run;
          %if &syserr^=0 %then %goto exittb;

          proc sort data=&outlib..TABLES force;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exittb;

          proc sort data=&tmplib..tables;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exittb;

          data &outlib..TABLES;
               merge &outlib..TABLES &tmplib..tables;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exittb;
      %end;
      %else %do;
          data &outlib..TABLES;
               set &tmplib..vtables;
          run;
          %if &syserr^=0 %then %goto exittb;

          proc sort data=&outlib..TABLES force;
               by language tabname;
          run;
          %if &syserr^=0 %then %goto exittb;

      %end;

      proc datasets lib=&outlib nolist;
           modify TABLES;
              index create TABNAME;
      quit;

%exittb:

%mend;
